import logging
import json
import uuid

import pymongo
import tornado.web

from ..base_handler import APIBase
from ..auth import authorization
from iceprod.server.util import nowstr

logger = logging.getLogger('rest.grids')


def setup(handler_cfg):
    """
    Setup method for Grids REST API.

    Args:
        handler_cfg (dict): args to pass to the route

    Returns:
        dict: routes, database, indexes
    """
    return {
        'routes': [
            (r'/grids', MultiGridsHandler, handler_cfg),
            (r'/grids/(?P<grid_id>\w+)', GridsHandler, handler_cfg),
        ],
        'database': 'grids',
        'indexes': {
            'grids': {
                'grid_id_index': {'keys': 'grid_id', 'unique': True},
            }
        }
    }


class MultiGridsHandler(APIBase):
    """
    Handle multi grids requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self):
        """
        Get grid entries.

        Returns:
            dict: {'uuid': {grid_data}}
        """
        ret = await self.db.grids.find(projection={'_id':False}).to_list(1000)
        self.write({row['grid_id']:row for row in ret})
        self.finish()

    @authorization(roles=['admin', 'system'])
    async def post(self):
        """
        Create a grid entry.

        Body should contain the grid data.

        Returns:
            dict: {'result': <grid_id>}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'host': str,
            'queues': dict,  # dict of {name: type}
            'version': str,  # iceprod version
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key {} should be of type {}'.format(k, req_fields[k])
                raise tornado.web.HTTPError(400, reason=r)

        # set some fields
        grid_id = uuid.uuid1().hex
        data['grid_id'] = grid_id
        data['start_date'] = nowstr()
        data['last_update'] = data['start_date']
        if 'debug' not in data:
            data['debug'] = False

        await self.db.grids.insert_one(data)
        self.set_status(201)
        self.write({'result': grid_id})
        self.finish()


class GridsHandler(APIBase):
    """
    Handle single grid requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, grid_id):
        """
        Get a grid entry.

        Args:
            grid_id (str): the grid id

        Returns:
            dict: grid entry
        """
        ret = await self.db.grids.find_one({'grid_id':grid_id}, projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Grid not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin', 'system'])
    async def patch(self, grid_id):
        """
        Update a grid entry.

        Body should contain the grid data to update.  Note that this will
        perform a merge (not replace).

        Args:
            grid_id (str): the grid id

        Returns:
            dict: updated grid entry
        """
        data = json.loads(self.request.body)
        if not data:
            raise tornado.web.HTTPError(400, reason='Missing update data')
        data['last_update'] = nowstr()

        ret = await self.db.grids.find_one_and_update(
            {'grid_id':grid_id},
            {'$set':data},
            projection={'_id':False},
            return_document=pymongo.ReturnDocument.AFTER
        )
        if not ret:
            self.send_error(404, reason="Grid not found")
        else:
            self.write(ret)
            self.finish()



"""
Functions relating to OpenSSL certificates.
"""

import os
import hashlib
import logging
from datetime import datetime
import uuid

from OpenSSL import SSL,crypto
from pyasn1.type import univ
from pyasn1.codec.der import encoder

from iceprod.core import functions

logger = logging.getLogger('ssl_cert')


def create_ca(cert_filename,key_filename,days=365,hostname=None):
    """Make a certificate authority and key pair"""
    cert_filename = os.path.abspath(os.path.expandvars(cert_filename))
    key_filename = os.path.abspath(os.path.expandvars(key_filename))
    logger.warning('making CA cert at %s',cert_filename)

    if not (os.path.exists(cert_filename)
            and os.path.exists(key_filename)):
        if hostname is None:
            # get hostname
            hostname = functions.gethostname().encode('utf-8')
            if hostname is None:
                raise Exception('Cannot get hostname')

        # create a key pair
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 4096)

        # create a self-signed cert
        cert = crypto.X509()
        cert.set_version(2)  # version 3, since count starts at 0
        cert.get_subject().C = "US"
        cert.get_subject().ST = "Wisconsin"
        cert.get_subject().L = "Madison"
        cert.get_subject().O = "University of Wisconsin-Madison"  # noqa: E741
        cert.get_subject().OU = "IceCube IceProd Root CA"
        cert.get_subject().CN = hostname
        cert.set_serial_number(1)
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(days*24*60*60)
        cert.set_issuer(cert.get_subject())
        cert.set_pubkey(k)

        # get integer public key for cert
        # pyOpenSSL doesn't provide a good access method,
        # so do it the hard way by extracting from key output
        pubkey = crypto.dump_privatekey(crypto.FILETYPE_TEXT, cert.get_pubkey())
        logger.info('%r',pubkey)
        pubkey = [x.strip() for x in pubkey.split(b'\n') if len(x) > 0 and x.startswith(b' ')]
        logger.info('%r',pubkey)
        pubkey = int(b''.join(pubkey).replace(b':',b''),16)

        # make asn1 DER encoding
        seq = univ.Sequence()
        seq.setComponentByPosition(0,univ.Integer(pubkey))
        seq.setComponentByPosition(1,univ.Integer(65537))
        enc = encoder.encode(seq)

        # get hash of DER
        hash = hashlib.sha1(enc).hexdigest().encode('utf-8')

        # add extensions
        cert.add_extensions([
            crypto.X509Extension(b"basicConstraints", True,
                                 b"CA:TRUE, pathlen:1"),
            crypto.X509Extension(b"keyUsage", True,
                                 b"keyCertSign, cRLSign"),
            crypto.X509Extension(b"subjectKeyIdentifier", False, hash,
                                 subject=cert),
            crypto.X509Extension(b"subjectAltName", False, b'DNS:'+hostname.encode('utf-8'))
        ])
        cert.sign(k, 'sha512')

        open(cert_filename, "wb").write(
            crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
        open(key_filename, "wb").write(
            crypto.dump_privatekey(crypto.FILETYPE_PEM, k))


def create_cert(cert_filename,key_filename,days=365,hostname=None,
                cacert=None,cakey=None,allow_resign=False):
    """Make a certificate and key pair"""
    cert_filename = os.path.abspath(os.path.expandvars(cert_filename))
    key_filename = os.path.abspath(os.path.expandvars(key_filename))
    logger.warning('making cert at %s',cert_filename)
    if cacert:
        cacert = os.path.abspath(os.path.expandvars(cacert))
        cakey = os.path.abspath(os.path.expandvars(cakey))
        logger.warning('with CA %s',cacert)
        if not (os.path.exists(cacert) and os.path.exists(cakey)):
            raise Exception('CA cert does not exist')

    if not (os.path.exists(cert_filename)
            and os.path.exists(key_filename)):
        if hostname is None:
            # get hostname
            hostname = functions.gethostname().encode('utf-8')
            if hostname is None:
                raise Exception('Cannot get hostname')

        # create a key pair
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 2048)

        if cacert is None or cakey is None:
            # self-signed
            cert = crypto.X509()
        else:
            # certificate request
            cert = crypto.X509Req()

        if cacert is None or cakey is None:
            # self-sign
            cert.get_subject().CN = hostname

            cert.gmtime_adj_notBefore(0)
            cert.gmtime_adj_notAfter(days*24*60*60)
            cert.set_serial_number(uuid.uuid4().int)
            cert.set_issuer(cert.get_subject())
            cert.set_pubkey(k)

            # add extensions
            exts = [crypto.X509Extension(b'subjectAltName', False, b'DNS:'+hostname.encode('utf-8'))]
            if allow_resign:
                exts.extend([
                    crypto.X509Extension("basicConstraints", True,
                                         "CA:TRUE, pathlen:0"),
                    # crypto.X509Extension("keyUsage", True,
                    #                      "keyCertSign, cRLSign"),
                    # crypto.X509Extension("subjectKeyIdentifier", False, hash,
                    #                      subject=cert),
                ])
            cert.add_extensions(exts)
            cert.sign(k, 'sha512')

        else:
            cert.get_subject().C = "US"
            cert.get_subject().ST = "Wisconsin"
            cert.get_subject().L = "Madison"
            cert.get_subject().O = "University of Wisconsin-Madison"  # noqa: E741
            cert.get_subject().OU = "IceCube IceProd"
            cert.get_subject().CN = hostname

            # finish cert req
            cert.set_pubkey(k)
            cert.sign(k, 'sha512')

            # load CA
            cacert = os.path.abspath(os.path.expandvars(cacert))
            cakey = os.path.abspath(os.path.expandvars(cakey))
            ca_cert = crypto.load_certificate(crypto.FILETYPE_PEM,open(cacert).read())
            ca_key = crypto.load_privatekey(crypto.FILETYPE_PEM,open(cakey).read())

            # make actual cert and sign with CA
            cert2 = crypto.X509()
            cert2.set_subject(cert.get_subject())
            cert2.set_serial_number(1)
            cert2.gmtime_adj_notBefore(0)
            cert2.gmtime_adj_notAfter(days*24*60*60)
            cert2.set_issuer(ca_cert.get_subject())
            cert2.set_pubkey(cert.get_pubkey())

            # add extensions
            exts = [crypto.X509Extension(b"subjectAltName", False, b'DNS:'+hostname.encode('utf-8'))]
            if allow_resign:
                exts.extend([
                    crypto.X509Extension(b"basicConstraints", True,
                                         b"CA:TRUE, pathlen:0"),
                    # crypto.X509Extension("keyUsage", True,
                    #                      "keyCertSign, cRLSign"),
                    # crypto.X509Extension("subjectKeyIdentifier", False, hash,
                    #                      subject=cert),
                ])
            cert2.add_extensions(exts)
            cert2.sign(ca_key, 'sha512')

            # overwrite cert req with real cert
            cert = cert2

        open(cert_filename, "wb").write(
            crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
        open(key_filename, "wb").write(
            crypto.dump_privatekey(crypto.FILETYPE_PEM, k))


def verify_cert(cert_filename, key_filename):
    """Verify if cert and key match.
       Return False for failure, True for success.
    """
    cert_filename = os.path.abspath(os.path.expandvars(cert_filename))
    key_filename = os.path.abspath(os.path.expandvars(key_filename))

    cert = crypto.load_certificate(crypto.FILETYPE_PEM,open(cert_filename).read())
    key = crypto.load_privatekey(crypto.FILETYPE_PEM,open(key_filename).read())

    # check date
    begin = cert.get_notBefore().decode('utf-8')
    logger.debug('begin: %r',begin)
    if datetime.strptime(begin, "%Y%m%d%H%M%SZ") > datetime.utcnow():
        logger.error('cert only valid in future')
        return False
    end = cert.get_notAfter().decode('utf-8')
    logger.debug('end: %r',end)
    if datetime.strptime(end, "%Y%m%d%H%M%SZ") < datetime.utcnow():
        logger.warning('cert has expired')
        return False

    # check matching cert and key
    ctx = SSL.Context(SSL.TLSv1_METHOD)
    ctx.use_privatekey(key)
    ctx.use_certificate(cert)
    try:
        ctx.check_privatekey()
    except SSL.Error:
        logger.warning('cert and key do not match')
        return False
    else:
        return True



"""The IceProd Grid Middleware and tools."""

from __future__ import absolute_import, division, print_function

# version is a human-readable version number.

# version_info is a four-tuple for programmatic comparison. The first
# three numbers are the components of the version number. The fourth
# is zero for an official release, positive for a development branch,
# or negative for a release candidate or beta (after the base version
# number has been incremented)
__version__ = "3.0.24"
version_info = (
    int(__version__.split(".")[0]),
    int(__version__.split(".")[1]),
    int(__version__.split(".")[2]),
    0,
)



"""
Some constants and basic functions with wide uses.
"""

from __future__ import absolute_import, division, print_function

import os
import sys
import logging
from contextlib import contextmanager
import tempfile

# some basic constants used for the task runner
constants = {'stdout': 'iceprod_out',
             'stderr': 'iceprod_err',
             'stdlog': 'iceprod_log',
             'stats': 'iceprod_stats',
             'task_exception': 'iceprod_task_exception',
             'args': 'iceprod_args',}


@contextmanager
def to_file(stream=None, file='out'):
    """redirect a stream (like stdout) to regular file

       :param stream: previous file object
       :param file: new file object
    """
    oldstream = os.dup(stream.fileno())
    with open(file,'w') as f:
        os.dup2(f.fileno(), stream.fileno())
        try:
            yield  # go do something useful
        finally:
            os.dup2(oldstream, stream.fileno())


@contextmanager
def to_log(stream=None, prefix='', level='info'):
    """redirect a stream (like stdout) to log file

       :param stream: file object
       :param prefix: logging prefix
       :param level: logging level
    """
    if not prefix:
        if stream == sys.stdout:
            prefix = 'stdout'
        elif stream == sys.stderr:
            prefix = 'stderr'
    logger = logging.getLogger(prefix)

    oldstream = os.dup(stream.fileno())
    with tempfile.TemporaryFile() as f:
        stream.flush()
        os.dup2(f.fileno(), stream.fileno())
        try:
            yield  # go do something useful
        finally:
            f.flush()
            os.dup2(oldstream, stream.fileno())
            f.seek(0)
            for s in f.readlines():
                s = s.strip()
                if not s:
                    continue
                if level == 'debug':
                    logger.debug(s)
                elif level == 'info':
                    logger.info(s)
                elif level in ('warn','warning'):
                    logger.warning(s)
                elif level in ('err','error'):
                    logger.error(s)
                elif level in ('crit','critical'):
                    logger.critical(s)
                else:
                    logger.warning('error printing logging')



import asyncio
from concurrent.futures import ThreadPoolExecutor
import io
from functools import partial
import logging

try:
    import boto3
    import botocore.client
    import botocore.exceptions
except ImportError:
    boto3 = None

logger = logging.getLogger('s3')


class S3:
    """S3 wrapper for uploading and downloading objects"""
    def __init__(self, address, access_key, secret_key, bucket='iceprod2-logs', mock_s3=None):
        self.bucket = bucket
        if mock_s3:
            self.s3 = mock_s3
        else:
            try:
                self.s3 = boto3.client(
                    's3',
                    'us-east-1',
                    endpoint_url=address,
                    aws_access_key_id=access_key,
                    aws_secret_access_key=secret_key,
                    config=botocore.client.Config(max_pool_connections=101)
                )
            except Exception:
                logger.warning('failed to connect to s3: %r', address, exc_info=True)
                raise
        self.executor = ThreadPoolExecutor(max_workers=20)

    async def create_bucket(self):
        """Create the bucket, if it does not exist"""
        def _inner():
            buckets = {b['Name'] for b in self.s3.list_buckets().get('Buckets', [])}
            if self.bucket not in buckets:
                self.s3.create_bucket(Bucket=self.bucket)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(self.executor, _inner)

    async def get(self, key):
        """Download object from S3"""
        ret = ''
        with io.BytesIO() as f:
            loop = asyncio.get_running_loop()
            try:
                await loop.run_in_executor(self.executor, partial(self.s3.download_fileobj, Bucket=self.bucket, Key=key, Fileobj=f))
                ret = f.getvalue()
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == '404':
                    return ''  # don't error on a 404
                raise
        return ret.decode('utf-8')

    def get_presigned(self, key, expiration=3600):
        """Make a presigned download url"""
        return self.s3.generate_presigned_url(
            'get_object',
            Params={'Bucket': self.bucket, 'Key': key},
            ExpiresIn=expiration,
        )

    async def put(self, key, data):
        """Upload object to S3"""
        with io.BytesIO(data.encode('utf-8')) as f:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(self.executor, partial(self.s3.upload_fileobj, f, self.bucket, key))

    def put_presigned(self, key, expiration=3600):
        """Make a presigned upload url"""
        return self.s3.generate_presigned_url(
            'put_object',
            Params={'Bucket': self.bucket, 'Key': key},
            ExpiresIn=expiration,
        )

    async def exists(self, key):
        """Check existence in S3"""
        loop = asyncio.get_running_loop()
        try:
            await loop.run_in_executor(self.executor, partial(self.s3.head_object, Bucket=self.bucket, Key=key))
        except Exception:
            return False
        return True

    async def delete(self, key):
        """Delete object in S3"""
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(self.executor, partial(self.s3.delete_object, Bucket=self.bucket, Key=key))

    async def list(self, prefix=None, recursive=True):
        """
        List bucket contents.

        Args:
            prefix (str): key prefix
            recursive (bool): recursive directory structure (default: True)
        Returns:
            dict: directory listing, with size for files
        """
        kwargs = {'Bucket': self.bucket}
        if prefix:
            kwargs['Prefix'] = prefix

        def _inner():
            paginator = self.s3.get_paginator('list_objects_v2')
            page_iterator = paginator.paginate(**kwargs)
            ret = {}
            for page in page_iterator:
                for item in page.get('Contents', []):
                    if not recursive:
                        ret[item['Key']] = item['Size']
                    else:
                        parts = item['Key'].split('/')
                        handle = ret
                        while len(parts) > 1:
                            d = parts.pop(0)
                            if d not in handle:
                                handle[d] = {}
                            handle = handle[d]
                        handle[parts[0]] = item['Size']
            return ret

        loop = asyncio.get_running_loop()
        ret = await loop.run_in_executor(self.executor, _inner)
        return ret

    async def rmtree(self, prefix):
        """Delete multiple objects in S3 by prefix"""
        objects = await self.list(prefix, recursive=False)
        loop = asyncio.get_running_loop()

        keys = [{'Key': k} for k in objects]
        while keys:
            d = {
                'Objects': keys[:1000],
                'Quiet': True,
            }
            keys = keys[1000:]
            await loop.run_in_executor(self.executor, partial(self.s3.delete_objects, Bucket=self.bucket, Delete=d))



"""
Credentials store and refresh.
"""
import asyncio
from datetime import datetime
import logging
import time

import pymongo
import pymongo.errors
import motor.motor_asyncio
import requests.exceptions
from rest_tools.client import RestClient, ClientCredentialsAuth
from rest_tools.server import RestServer, ArgumentHandler, ArgumentSource
from tornado.web import HTTPError
from tornado.web import RequestHandler as TornadoRequestHandler
from wipac_dev_tools import from_environment

from iceprod.rest.auth import authorization
from iceprod.rest.base_handler import IceProdRestConfig, APIBase
from iceprod.server.module import FakeStatsClient, StatsClientIgnoreErrors
from iceprod.server.util import nowstr, datetime2str
from .service import RefreshService, get_expiration, is_expired

logger = logging.getLogger('server')


class BaseCredentialsHandler(APIBase):
    def initialize(self, refresh_service=None, rest_client=None, **kwargs):
        super().initialize(**kwargs)
        self.refresh_service = refresh_service
        self.rest_client = rest_client

    async def check_attr_auth(self, arg, val, role):
        """
        Based on the request groups or username, check if they are allowed to
        access `arg`:`role`.

        Runs a remote query to the IceProd API.

        Args:
            arg (str): attribute name to check
            val (str): attribute value
            role (str): the role to check for (read|write)
        """
        args = {
            'name': arg,
            'value': val,
            'role': role,
            'username': self.current_user,
            'groups': self.auth_groups,
        }
        try:
            await self.rest_client.request('POST', '/auths', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 403:
                raise HTTPError(403, 'auth failed')
            else:
                raise HTTPError(500, 'auth could not be completed')

    async def create(self, db, base_data):
        now = time.time()
        argo = ArgumentHandler(ArgumentSource.JSON_BODY_ARGUMENTS, self)
        argo.add_argument('url', type=str, required=True)
        argo.add_argument('type', type=str, choices=['s3', 'oauth'], required=True)
        argo.add_argument('buckets', type=list, default=[], required=False)
        argo.add_argument('access_key', type=str, default='', required=False)
        argo.add_argument('secret_key', type=str, default='', required=False)
        argo.add_argument('access_token', type=str, default='', required=False)
        argo.add_argument('refresh_token', type=str, default='', required=False)
        argo.add_argument('expire_date', type=float, default=now, required=False)
        argo.add_argument('last_use', type=float, default=now, required=False)
        args = vars(argo.parse_args())
        url = args['url']
        credential_type = args['type']

        base_data['url'] = url
        data = base_data.copy()
        data.update({
            'type': credential_type,
        })

        if credential_type == 's3':
            if not args['buckets']:
                raise HTTPError(400, reason='must specify bucket(s)')
            if not args['access_key']:
                raise HTTPError(400, reason='must specify access_key')
            if not args['secret_key']:
                raise HTTPError(400, reason='must specify secret_key')

            data['buckets'] = args['buckets']
            data['access_key'] = args['access_key']
            data['secret_key'] = args['secret_key']

        elif credential_type == 'oauth':
            if (not args['access_token']) and not args['refresh_token']:
                raise HTTPError(400, reason='must specify either access or refresh tokens')
            data['access_token'] = args['access_token']
            data['refresh_token'] = args['refresh_token']
            data['expiration'] = args['expire_date']
            data['last_use'] = args['last_use']

            if 'refresh_token' in data and not data.get('access_token', ''):
                new_cred = await self.refresh_service.refresh_cred(data)
                data.update(new_cred)
            if (not data.get('access_token', '')) and data.get('expiration') == now:
                data['expiration'] = get_expiration(data['access_token'])

        else:
            raise HTTPError(400, 'bad credential type')

        await db.update_one(
            base_data,
            {'$set': data},
            upsert=True,
        )

    async def patch_cred(self, db, base_data):
        argo = ArgumentHandler(ArgumentSource.JSON_BODY_ARGUMENTS, self)
        argo.add_argument('url', type=str, required=True)
        argo.add_argument('buckets', type=list, default=[], required=False)
        argo.add_argument('access_key', type=str, default='', required=False)
        argo.add_argument('secret_key', type=str, default='', required=False)
        argo.add_argument('access_token', type=str, default='', required=False)
        argo.add_argument('refresh_token', type=str, default='', required=False)
        argo.add_argument('expiration', type=float, default=0, required=False)
        argo.add_argument('last_use', type=float, default=0, required=False)
        args = vars(argo.parse_args())
        base_data['url'] = args['url']

        data = {}
        for key in ('buckets', 'access_key', 'secret_key', 'access_token', 'refresh_token', 'expiration', 'last_use'):
            if val := args[key]:
                data[key] = val

        if 'refresh_token' in data and 'access_token' not in data:
            new_cred = await self.refresh_service.refresh_cred(data)
            data.update(new_cred)
        if 'access_token' in data and 'expiration' not in data:
            data['expiration'] = get_expiration(data['access_token'])

        ret = await db.find_one_and_update(
            base_data,
            {'$set': data},
        )
        if not ret:
            raise HTTPError(404, 'credential not found')

    async def search_creds(self, db, base_data):
        if url := self.get_argument('url', None):
            base_data['url'] = url

        refresh = self.get_argument('norefresh', None) is None

        if refresh:
            update_data = {'last_use': time.time()}
            filters = base_data.copy()
            filters['type'] = 'oauth'
            await db.update_many(filters, {'$set': update_data})

        ret = {}
        async for row in db.find(base_data, projection={'_id': False}):
            ret[row['url']] = row

        for key in list(ret):
            cred = ret[key]
            if refresh and is_expired(cred) and cred['refresh_token']:
                try:
                    new_cred = await self.refresh_service.refresh_cred(cred)
                    filters = base_data.copy()
                    filters['url'] = key
                    ret[key] = await db.find_one_and_update(filters, {'$set': new_cred}, projection={'_id': False})
                except Exception:
                    del ret[key]

        return ret


class GroupCredentialsHandler(BaseCredentialsHandler):
    """
    Handle group credentials requests.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def get(self, groupname):
        """
        Get a group's credentials.

        Args:
            groupname (str): group name
        Returns:
            dict: url: credential dict
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        ret = await self.search_creds(self.db.group_creds, {'groupname': groupname})
        self.write(ret)

    @authorization(roles=['admin', 'system', 'user'])
    async def post(self, groupname):
        """
        Set a group credential.  Overwrites an existing credential for the specified url.

        Common body args:
            url (str): url of controlled resource
            type (str): credential type (`s3` or `oauth`)

        S3 body args:
            buckets (list): list of buckets for this url, or [] if using virtual-hosted buckets in the url
            access_key (str): access key
            secret_key (str): secret key

        OAuth body args:
            access_token (str): access token
            refresh_token (str): refresh token
            expire_date (str): access token expiration, ISO date time in UTC (optional)

        Args:
            groupname (str): group name
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        await self.create(self.db.group_creds, {'groupname': groupname})
        self.write({})

    @authorization(roles=['admin', 'system', 'user'])
    async def patch(self, groupname):
        """
        Update a group credential.  Usually used to update a specifc field.

        Required body args:
            url (str): url of controlled resource

        Other body args will update a credential.

        Args:
            groupname (str): group name
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        await self.patch_cred(self.db.group_creds, {'groupname': groupname})
        self.write({})

    @authorization(roles=['admin', 'system', 'user'])
    async def delete(self, groupname):
        """
        Delete a group's credentials.

        Args:
            groupname (str): groupname
        Body args:
            url (str): (optional) url of controlled resource
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        args = {'groupname': groupname}

        argo = ArgumentHandler(ArgumentSource.JSON_BODY_ARGUMENTS, self)
        argo.add_argument('url', type=str, default='', required=False)
        body_args = argo.parse_args()
        if body_args.url:
            args['url'] = body_args.url

        await self.db.group_creds.delete_many(args)
        self.write({})


class UserCredentialsHandler(BaseCredentialsHandler):
    """
    Handle user credentials requests.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def get(self, username):
        """
        Get a user's credentials.

        Args:
            username (str): username
        Returns:
            dict: url: credential dict
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        ret = await self.search_creds(self.db.user_creds, {'username': username})
        self.write(ret)

    @authorization(roles=['admin', 'system', 'user'])
    async def post(self, username):
        """
        Set a user credential.  Overwrites an existing credential for the specified url.

        Common body args:
            url (str): url of controlled resource
            type (str): credential type (`s3` or `oauth`)

        S3 body args:
            buckets (list): list of buckets for this url, or [] if using virtual-hosted buckets in the url
            access_key (str): access key
            secret_key (str): secret key

        OAuth body args:
            access_token (str): access token
            refresh_token (str): refresh token
            expire_date (str): access token expiration, ISO date time in UTC (optional)

        Args:
            username (str): username
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        await self.create(self.db.user_creds, {'username': username})
        self.write({})

    @authorization(roles=['admin', 'system', 'user'])
    async def patch(self, username):
        """
        Update a user credential.  Usually used to update a specifc field.

        Required body args:
            url (str): url of controlled resource

        Other body args will update a credential.

        Args:
            username (str): username
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        await self.patch_cred(self.db.user_creds, {'username': username})
        self.write({})

    @authorization(roles=['admin', 'system', 'user'])
    async def delete(self, username):
        """
        Delete a user's credentials.

        Args:
            username (str): username
        Body args:
            url (str): (optional) url of controlled resource
        Returns:
            dict: url: credential dict
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        args = {'username': username}

        argo = ArgumentHandler(ArgumentSource.JSON_BODY_ARGUMENTS, self)
        argo.add_argument('url', type=str, default='', required=False)
        body_args = argo.parse_args()
        if body_args.url:
            args['url'] = body_args.url

        await self.db.user_creds.delete_many(args)
        self.write({})


class HealthHandler(BaseCredentialsHandler):
    """
    Handle health requests.
    """
    async def get(self):
        """
        Get health status.

        Returns based on exit code, 200 = ok, 400 = failure
        """
        now = time.time()
        status = {
            'now': nowstr(),
            'start_time': datetime2str(datetime.utcfromtimestamp(self.refresh_service.start_time)),
            'last_run_time': "",
            'last_success_time': "",
        }
        try:
            if self.refresh_service.last_run_time is None and self.refresh_service.start_time + 3600 < now:
                self.send_error(500, reason='refresh was never run')
                return
            if self.refresh_service.last_run_time is not None:
                if self.refresh_service.last_run_time + 3600 < now:
                    self.send_error(500, reason='refresh has stopped running')
                    return
                status['last_run_time'] = datetime2str(datetime.utcfromtimestamp(self.refresh_service.last_run_time))
            if self.refresh_service.last_success_time is None and self.refresh_service.start_time + 86400 < now:
                self.send_error(500, reason='refresh was never successful')
                return
            if self.refresh_service.last_success_time is not None:
                if self.refresh_service.last_success_time + 86400 < now:
                    self.send_error(500, reason='refresh has stopped being successful')
                    return
                status['last_success_time'] = datetime2str(datetime.utcfromtimestamp(self.refresh_service.last_success_time))
        except Exception:
            logger.info('error from refresh service', exc_info=True)
            self.send_error(500, reason='error from refresh service')
            return

        self.write(status)


class Error(TornadoRequestHandler):
    def prepare(self):
        raise HTTPError(404, 'invalid api route')


class Server:
    def __init__(self):
        default_config = {
            'HOST': 'localhost',
            'PORT': 8080,
            'DEBUG': False,
            'OPENID_URL': '',
            'OPENID_AUDIENCE': '',
            'ICEPROD_API_ADDRESS': 'https://iceprod2-api.icecube.wisc.edu',
            'ICEPROD_API_CLIENT_ID': '',
            'ICEPROD_API_CLIENT_SECRET': '',
            'TOKEN_CLIENTS': '{}',
            'TOKEN_REFRESH_WINDOW': 72.0,
            'TOKEN_EXPIRE_BUFFER': 24.0,
            'TOKEN_SERVICE_CHECK_INTERVAL': 180,
            'DB_URL': 'mongodb://localhost/creds',
            'DB_TIMEOUT': 60,
            'DB_WRITE_CONCERN': 1,
            'STATSD_ADDRESS': '',
            'STATSD_PREFIX': 'credentials',
            'CI_TESTING': '',
        }
        config = from_environment(default_config)

        rest_config = {
            'debug': config['DEBUG'],
        }
        if config['OPENID_URL']:
            logging.info(f'enabling auth via {config["OPENID_URL"]} for aud "{config["OPENID_AUDIENCE"]}"')
            rest_config.update({
                'auth': {
                    'openid_url': config['OPENID_URL'],
                    'audience': config['OPENID_AUDIENCE'],
                }
            })
        elif config['CI_TESTING']:
            rest_config.update({
                'auth': {
                    'secret': 'secret',
                }
            })
        else:
            raise RuntimeError('OPENID_URL not specified, and CI_TESTING not enabled!')

        statsd = FakeStatsClient()
        if config['STATSD_ADDRESS']:
            try:
                addr = config['STATSD_ADDRESS']
                port = 8125
                if ':' in addr:
                    addr,port = addr.split(':')
                    port = int(port)
                statsd = StatsClientIgnoreErrors(addr, port=port, prefix=config['STATSD_PREFIX'])
            except Exception:
                logger.warning('failed to connect to statsd: %r', config['STATSD_ADDRESS'], exc_info=True)

        logging_url = config["DB_URL"].split('@')[-1] if '@' in config["DB_URL"] else config["DB_URL"]
        logging.info(f'DB: {logging_url}')
        db_url, db_name = config['DB_URL'].rsplit('/', 1)
        db = motor.motor_asyncio.AsyncIOMotorClient(
            db_url,
            timeoutMS=config['DB_TIMEOUT']*1000,
            w=config['DB_WRITE_CONCERN'],
        )
        logging.info(f'DB name: {db_name}')
        self.db = db[db_name]
        self.indexes = {
            'group_creds': {
                'group_index': {'keys': [('groupname', pymongo.DESCENDING), ('url', pymongo.DESCENDING)], 'unique': True},
            },
            'user_creds': {
                'username_index': {'keys': [('username', pymongo.DESCENDING), ('url', pymongo.DESCENDING)], 'unique': True},
            }
        }

        if config['ICEPROD_API_CLIENT_ID'] and config['ICEPROD_API_CLIENT_SECRET']:
            logging.info(f'enabling auth via {config["OPENID_URL"]} for aud "{config["OPENID_AUDIENCE"]}"')
            rest_client = ClientCredentialsAuth(
                address=config['ICEPROD_API_ADDRESS'],
                token_url=config['OPENID_URL'],
                client_id=config['ICEPROD_API_CLIENT_ID'],
                client_secret=config['ICEPROD_API_CLIENT_SECRET'],
            )
        elif config['CI_TESTING']:
            rest_client = RestClient(config['ICEPROD_API_ADDRESS'], timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_API_CLIENT_ID or ICEPROD_API_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        self.refresh_service = RefreshService(
            database=self.db,
            clients=config['TOKEN_CLIENTS'],
            refresh_window=config['TOKEN_REFRESH_WINDOW'],
            expire_buffer=config['TOKEN_EXPIRE_BUFFER'],
            service_run_interval=config['TOKEN_SERVICE_CHECK_INTERVAL'],
        )
        self.refresh_service_task = None

        kwargs = IceProdRestConfig(rest_config, statsd=statsd, database=self.db)
        kwargs['refresh_service'] = self.refresh_service
        kwargs['rest_client'] = rest_client

        server = RestServer(debug=config['DEBUG'])

        server.add_route(r'/groups/(?P<groupname>\w+)/credentials', GroupCredentialsHandler, kwargs)
        server.add_route(r'/users/(?P<username>\w+)/credentials', UserCredentialsHandler, kwargs)
        server.add_route('/healthz', HealthHandler, kwargs)
        server.add_route(r'/(.*)', Error)

        server.startup(address=config['HOST'], port=config['PORT'])

        self.server = server

    async def start(self):
        for collection in self.indexes:
            existing = await self.db[collection].index_information()
            for name in self.indexes[collection]:
                if name not in existing:
                    logging.info('DB: creating index %s:%s', collection, name)
                    kwargs = self.indexes[collection][name]
                    await self.db[collection].create_index(name=name, **kwargs)

        if not self.refresh_service_task:
            self.refresh_service_task = asyncio.create_task(self.refresh_service.run())

    async def stop(self):
        await self.server.stop()
        if self.refresh_service_task:
            self.refresh_service_task.cancel()
            try:
                await self.refresh_service_task
            except asyncio.CancelledError:
                pass  # ignore cancellations
            finally:
                self.refresh_service_task = None



import logging
import json

import pymongo.errors
import tornado.web

from ..base_handler import APIBase
from ..auth import authorization
from iceprod.roles_groups import ROLES, GROUP_PRIORITIES

logger = logging.getLogger('rest.auth')


def setup(handler_cfg):
    """
    Setup method for Config REST API.

    Args:
        handler_cfg (dict): args to pass to the route

    Returns:
        dict: routes, database, indexes
    """
    return {
        'routes': [
            ('/roles', MultiRoleHandler, handler_cfg),
            ('/groups', MultiGroupHandler, handler_cfg),
            ('/users', MultiUserHandler, handler_cfg),
            (r'/users/(?P<username>[^\/\?\#]+)', UserHandler, handler_cfg),
            (r'/users/(?P<username>[^\/\?\#]+)/priority', UserPriorityHandler, handler_cfg),
            ('/auths', AuthHandler, handler_cfg),
        ],
        'database': 'auth',
        'indexes': {
            'users': {
                'username_index': {'keys': 'username', 'unique': True},
            },
            'attr_auths': {
                'dataset_id_index': {'keys': 'dataset_id', 'unique': False},
            }
        }
    }


class MultiRoleHandler(APIBase):
    """
    Handle multi-role requests.
    """
    @authorization(roles=['admin'])
    async def get(self):
        """
        Get a list of roles.

        Returns:
            dict: {'results': list of roles}
        """
        self.write({'results': list(ROLES)})


class MultiGroupHandler(APIBase):
    """
    Handle multi-group requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self):
        """
        Get a list of groups.

        Returns:
            dict: {'results': list of groups}
        """
        ret = []
        for name, priority in GROUP_PRIORITIES.items():
            ret.append({
                'name': name,
                'priority': priority,
            })
        self.write({'results': ret})


class MultiUserHandler(APIBase):
    """
    Handle multi-user requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self):
        """
        Get a list of users.

        Returns:
            dict: {'results': list of users}
        """
        ret = await self.db.users.find(projection={'_id': False}).to_list(length=10000)
        self.write({'results': ret})
        self.finish()


class UserHandler(APIBase):
    """
    Handle individual user requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, username):
        """
        Get a user.

        Args:
            username (str): the user to get
        Returns:
            dict: user info
        """
        ret = await self.db.users.find_one({'username': username}, projection={'_id':False})
        if not ret:
            self.send_error(404, reason="User not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin'])
    async def put(self, username):
        """
        Add a User.

        Args:
            username (str): the user to delete
        Returns:
            dict: empty dict
        """
        try:
            await self.add_user(username)
        except pymongo.errors.DuplicateKeyError:
            raise tornado.web.HTTPError(409, reason='duplicate username')
        self.write({})
        self.finish()

    @authorization(roles=['admin'])
    async def delete(self, username):
        """
        Delete a user.

        Args:
            username (str): the user to delete
        Returns:
            dict: empty dict
        """
        await self.db.users.delete_one({'username': username})
        self.write({})
        self.finish()


class UserPriorityHandler(APIBase):
    """
    Handle individual user requests.
    """
    @authorization(roles=['admin'])
    async def put(self, username):
        """
        Add a User.

        Args:
            username (str): the user to modify
        Body Args:
            priority (float): the new priority
        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'priority' not in data:
            raise tornado.web.HTTPError(400, reason='missing priority in body')
        try:
            priority = float(data['priority'])
            if not 0 <= priority <= 1:
                raise ValueError()
        except Exception:
            raise tornado.web.HTTPError(400, reason='bad priority: should be between 0 and 1')

        ret = await self.db.users.update_one({'username': username}, {'$set': {'priority': priority}})
        if ret.matched_count < 1:
            raise tornado.web.HTTPError(400, reason='bad username')
        self.write({})
        self.finish()


class AuthHandler(APIBase):
    """
    Handle authorization requests.
    """
    @authorization(roles=['admin', 'system'])
    async def post(self):
        """
        Do a remote auth lookup.  Raises a 403 on auth failure.

        Body Args:
            name (str): name of attr
            value (str): value of attr
            role (str): the role to check (read | write)
            username (str): username
            groups (list): groups for user

        Returns:
            dict: {result: ok}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'name': str,
            'value': str,
            'role': str,
            'username': str,
            'groups': list,
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key "{}" should be of type {}'.format(k, req_fields[k].__name__)
                raise tornado.web.HTTPError(400, reason=r)

        # check auth
        self.current_user = data['username']
        self.auth_groups = data['groups']
        await self.check_attr_auth(data['name'], data['value'], data['role'])
        self.write({})



"""
The task runner.

Run it with `python -m iceprod.core.i3exec`.
"""

import argparse
import asyncio
import json
import logging
from pathlib import Path
import subprocess

import iceprod
from iceprod.client_auth import add_auth_to_argparse, create_rest_client
import iceprod.core.config
import iceprod.core.exe
import iceprod.core.logger


logger = logging.getLogger('i3exec')


async def run(args):
    if args.dataset_id:
        rc = create_rest_client(args)
        logger.info('Real dataset mode: dataset %s task %s', args.dataset_id, args.task_id)
        task = await iceprod.core.config.Task.load_from_api(args.dataset_id, args.task_id, rc)

    else:
        logger.info('Testing mode: dataset %d job %d task %s', args.dataset_num, args.job_index, args.task)
        with open(args.config) as f:
            cfg = json.load(f)
        task_names = [t['name'] for t in cfg['tasks']]

        d = iceprod.core.config.Dataset(
            dataset_id='datasetid',
            dataset_num=args.dataset_num,
            jobs_submitted=args.jobs_submitted,
            tasks_submitted=args.jobs_submitted*len(cfg['tasks']),
            tasks_per_job=len(cfg['tasks']),
            status='processing',
            priority=0,
            group='group',
            user='user',
            debug=True,
            config=cfg
        )
        task = iceprod.core.config.Task(
            dataset=d,
            job=iceprod.core.config.Job(d, '', args.job_index, 'processing'),
            task_id='taskid',
            task_index=task_names.index(args.task),
            name=args.task,
            depends=[],
            requirements={},
            status='processing',
            site='site',
            stats={}
        )

    task.dataset.fill_defaults()
    task.dataset.validate()

    ws = iceprod.core.exe.WriteToScript(task, workdir=Path.cwd(), logger=logger)
    scriptpath = await ws.convert()
    logger.info('running script %s', scriptpath)
    if not args.dry_run:
        subprocess.run([scriptpath], check=True)


async def main():
    parser = argparse.ArgumentParser(description='IceProd Core')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('-n', '--dry-run', action='store_true', default=False, help='Dry run')
    add_auth_to_argparse(parser)

    real = parser.add_argument_group('Real Dataset', 'Download from IceProd server')
    real.add_argument('--dataset-id', help='IceProd dataset id')
    real.add_argument('--task-id', help='IceProd task id')

    testing = parser.add_argument_group('Testing')
    testing.add_argument('--config', help='Specify config file')
    testing.add_argument('--task', type=str, help='Name of the task to run')
    testing.add_argument('--dataset-num', type=int, default=1, help='Fake dataset number (optional)')
    testing.add_argument('--jobs-submitted', type=int, default=1, help='Total number of jobs in this dataset (optional)')
    testing.add_argument('--job-index', type=int, default=0, help='Fake job index (optional)')

    args = parser.parse_args()

    if args.dataset_id:
        if not args.task_id:
            parser.error('task-id is required')
    else:
        if not args.config:
            parser.error('config is required')
        if not args.task:
            parser.error('task is required')

    iceprod.core.logger.set_logger(loglevel=args.log_level)

    await run(args)


if __name__ == '__main__':
    asyncio.run(main())



"""
Main website
------------

This is the external website users will see when interacting with IceProd.
It has been broken down into several sub-handlers for easier maintenance.
"""

from collections import defaultdict
from datetime import datetime, timedelta
import functools
import importlib.resources
import logging
import os
import random
import re
from urllib.parse import urlencode

from iceprod.core.jsonUtil import json_encode

from cachetools.func import ttl_cache
import tornado.web
import tornado.httpserver
import tornado.gen
import jwt
import tornado.concurrent
from rest_tools.client import RestClient, ClientCredentialsAuth
from rest_tools.server import catch_error, RestServer, RestHandlerSetup, RestHandler, OpenIDLoginHandler
from rest_tools import telemetry as wtt
from wipac_dev_tools import from_environment

import iceprod
from iceprod.roles_groups import GROUPS
from iceprod.core.config import CONFIG_SCHEMA as DATASET_SCHEMA
from iceprod.server.config import CONFIG_SCHEMA as SERVER_SCHEMA
import iceprod.core.functions
from iceprod.server import documentation
from iceprod.server.module import FakeStatsClient, StatsClientIgnoreErrors
import iceprod.server.states
from iceprod.server.util import nowstr

logger = logging.getLogger('website')


def authenticated(method):
    """Decorate methods with this to require that the user be logged in.

    If the user is not logged in, they will be redirected to the configured
    `login url <RequestHandler.get_login_url>`.

    If you configure a login url with a query parameter, Tornado will
    assume you know what you're doing and use it as-is.  If not, it
    will add a `next` parameter so the login page knows where to send
    you once you're logged in.
    """
    @catch_error
    @functools.wraps(method)
    async def wrapper(self, *args, **kwargs):
        if not await self.get_current_user_async():
            if self.request.method in ("GET", "HEAD"):
                try:
                    url = self.get_login_url()
                    if "?" not in url:
                        next_url = self.request.uri
                        url += "?" + urlencode(dict(next=next_url))
                    self.redirect(url)
                    return None
                except Exception:
                    logger.warning('failed to make redirect', exc_info=True)
                    raise tornado.web.HTTPError(403, reason='auth failed')
            raise tornado.web.HTTPError(403, reason='auth failed')
        return await method(self, *args, **kwargs)
    return wrapper


def eval_expression(token, e):
    """from rest_tools.server.decorators.token_attribute_role_mapping_auth"""
    name, val = e.split('=',1)
    if name == 'scope':
        # special handling to split into string
        token_val = token.get('scope','').split()
    else:
        prefix = name.split('.')[:-1]
        while prefix:
            token = token.get(prefix[0], {})
            prefix = prefix[1:]
        token_val = token.get(name.split('.')[-1], None)

    logger.debug('token_val = %r', token_val)
    if token_val is None:
        return []

    prog = re.compile(val)
    if isinstance(token_val, list):
        ret = (prog.fullmatch(v) for v in token_val)
    else:
        ret = [prog.fullmatch(token_val)]
    return [r for r in ret if r]


class TokenStorageMixin:
    """
    Store/load current user's `OpenIDLoginHandler` tokens in iceprod credentials API.
    """
    def get_current_user(self):
        return None

    async def get_current_user_async(self):
        """Get the current user, and set auth-related attributes."""
        try:
            username = self.get_secure_cookie('iceprod_username')
            if not username:
                return None
            if isinstance(username, bytes):
                username = username.decode('utf-8')
            creds = await self.cred_rest_client.request('GET', f'/users/{username}/credentials', {'url': self.full_url})
            cred = creds[self.full_url]
            access_token = cred['access_token']
            try:
                data = self.auth.validate(access_token)
            except jwt.ExpiredSignatureError:
                logger.debug('user access_token expired')
                return None
            self.auth_data = data

            # lookup groups
            auth_groups = set()
            try:
                for name in GROUPS:
                    for expression in GROUPS[name]:
                        ret = eval_expression(data, expression)
                        auth_groups.update(match.expand(name) for match in ret)
            except Exception:
                logger.info('cannot determine groups', exc_info=True)
            self.auth_groups = sorted(auth_groups)

            self.auth_access_token = access_token
            self.auth_refresh_token = cred.get('refresh_token', '')
            return username

        except Exception:
            logger.debug('failed auth', exc_info=True)
        return None

    def store_tokens(
        self,
        access_token,
        access_token_exp,
        refresh_token=None,
        refresh_token_exp=None,
        user_info=None,
        user_info_exp=None,
    ):
        """
        Store jwt tokens and user info from OpenID-compliant auth source.

        Args:
            access_token (str): jwt access token
            access_token_exp (int): access token expiration in seconds
            refresh_token (str): jwt refresh token
            refresh_token_exp (int): refresh token expiration in seconds
            user_info (dict): user info (from id token or user info lookup)
            user_info_exp (int): user info expiration in seconds
        """
        if not user_info:
            user_info = self.auth.validate(access_token)
        username = user_info.get('preferred_username')
        if not username:
            username = user_info.get('upn')
        if not username:
            raise tornado.web.HTTPError(400, reason='no username in token')
        args = {
            'url': self.full_url,
            'type': 'oauth',
            'access_token': access_token,
        }
        if refresh_token:
            args['refresh_token'] = refresh_token

        self.cred_rest_client.request_seq('POST', f'/users/{username}/credentials', args)

        self.set_secure_cookie('iceprod_username', username, expires_days=30)

    def clear_tokens(self):
        """
        Clear token data, usually on logout.
        """
        self.clear_cookie('iceprod_username')


class Login(TokenStorageMixin, OpenIDLoginHandler):
    def initialize(self, cred_rest_client=None, full_url=None, **kwargs):
        super().initialize(**kwargs)
        self.cred_rest_client = cred_rest_client
        self.full_url = full_url


class PublicHandler(TokenStorageMixin, RestHandler):
    """Default Handler"""
    def initialize(self, statsd=None, rest_api=None, cred_rest_client=None, system_rest_client=None, full_url=None, **kwargs):
        """
        Get some params from the website module

        :param statsd: statsd client
        :param rest_api: the rest api url
        :param cred_rest_client: the rest api url for the cred service
        :param system_rest_client: the rest client for the system role
        :param full_url: the full base url for the website
        """
        super().initialize(**kwargs)
        self.statsd = statsd
        self.rest_api = rest_api
        self.cred_rest_client = cred_rest_client
        self.system_rest_client = system_rest_client
        self.rest_client = None
        self.full_url = full_url

    def get_template_namespace(self):
        namespace = super().get_template_namespace()
        namespace['version'] = iceprod.__version__
        namespace['section'] = self.request.uri.lstrip('/').split('?')[0].split('/')[0]
        namespace['json_encode'] = json_encode
        namespace['states'] = iceprod.server.states
        namespace['rest_api'] = self.rest_api
        return namespace

    async def get_current_user_async(self):
        ret = await super().get_current_user_async()
        try:
            if ret:
                self.rest_client = RestClient(self.rest_api, self.auth_access_token, timeout=50, retries=1)
        except Exception:
            pass

        self.current_user = ret
        return ret

    @wtt.evented(all_args=True)
    def write_error(self, status_code=500, **kwargs):
        """Write out custom error page."""
        self.set_status(status_code)
        if status_code >= 500:
            self.write('<h2>Internal Error</h2>')
        else:
            self.write('<h2>Request Error</h2>')
        if 'message' in kwargs:
            self.write('<br />'.join(kwargs['message'].split('\n')))
        elif 'reason' in kwargs:
            self.write('<br />'.join(kwargs['reason'].split('\n')))
        elif self._reason:
            self.write('<br />'.join(self._reason.split('\n')))
        self.finish()


class Default(PublicHandler):
    """Handle / urls"""
    @catch_error
    async def get(self):
        # try to get the user, if available
        await self.get_current_user_async()
        self.statsd.incr('default')
        self.render('main.html')


class Schemas(PublicHandler):
    """Handle /schemas/v3/(.*) urls"""
    @catch_error
    async def get(self, schema):
        if schema == 'dataset.schema.json':
            self.write(DATASET_SCHEMA)
        elif schema == 'config.schema.json':
            self.write(SERVER_SCHEMA)
        else:
            raise tornado.web.HTTPError(404, reason='unknown schema')


class Submit(PublicHandler):
    """Handle /submit urls"""
    @authenticated
    async def get(self):
        logger.info('here')
        self.statsd.incr('submit')
        token = self.auth_access_token
        groups = self.auth_groups
        default_config = {
            "categories": [],
            "dataset": 0,
            "description": "",
            "difplus": None,
            "options": {},
            "parent_id": 0,
            "steering": None,
            "tasks": [],
            "version": 3
        }
        render_args = {
            'passkey':token,
            'edit':False,
            'dataset_id':'',
            'config':default_config,
            'groups':groups,
            'description':'',
        }
        self.render('submit.html', **render_args)


class Config(PublicHandler):
    """Handle /config urls"""
    @authenticated
    async def get(self):
        self.statsd.incr('config')
        dataset_id = self.get_argument('dataset_id',default=None)
        if not dataset_id:
            self.write_error(400,message='must provide dataset_id')
            return
        dataset = await self.rest_client.request('GET','/datasets/{}'.format(dataset_id))
        edit = self.get_argument('edit',default=False)
        if edit:
            passkey = self.auth_access_token
        else:
            passkey = None
        config = await self.rest_client.request('GET','/config/{}'.format(dataset_id))
        render_args = {
            'edit':edit,
            'passkey':passkey,
            'dataset_id':dataset_id,
            'config':config,
            'description':dataset.get('description',''),
        }
        self.render('submit.html',**render_args)


class DatasetBrowse(PublicHandler):
    """Handle /dataset urls"""
    @ttl_cache(maxsize=256, ttl=600)
    async def get_usernames(self):
        ret = await self.system_rest_client.request('GET', '/users')
        return [x['username'] for x in ret['results']]

    @authenticated
    async def get(self):
        self.statsd.incr('dataset_browse')
        usernames = await self.get_usernames()
        filter_options = {
            'status': ['processing', 'suspended', 'errors', 'complete', 'truncated'],
            'groups': list(GROUPS.keys()),
            'users': usernames,
        }
        filter_results = {n: self.get_arguments(n) for n in filter_options}

        args = {'keys': 'dataset_id|dataset|status|description|group|username'}
        for name in filter_results:
            val = filter_results[name]
            if not val:
                continue
            if any(v not in filter_options[name] for v in val):
                raise tornado.web.HTTPError(400, reason='Bad filter '+name+' value')
            args[name] = '|'.join(val)

        url = '/datasets'

        ret = await self.rest_client.request('GET', url, args)
        datasets = sorted(ret.values(), key=lambda x:x.get('dataset',0), reverse=True)
        logger.debug('datasets: %r', datasets)
        datasets = filter(lambda x: 'dataset' in x, datasets)
        self.render(
            'dataset_browse.html',
            datasets=datasets,
            filter_options=filter_options,
            filter_results=filter_results,
        )


class Dataset(PublicHandler):
    """Handle /dataset urls"""
    @authenticated
    async def get(self, dataset_id):
        self.statsd.incr('dataset')

        if dataset_id.isdigit():
            try:
                d_num = int(dataset_id)
                if d_num < 10000000:
                    all_datasets = await self.rest_client.request('GET', '/datasets', {'keys': 'dataset_id|dataset'})
                    for d in all_datasets.values():
                        if d['dataset'] == d_num:
                            dataset_id = d['dataset_id']
                            break
            except Exception:
                pass
        try:
            dataset = await self.rest_client.request('GET','/datasets/{}'.format(dataset_id))
        except Exception:
            raise tornado.web.HTTPError(404, reason='Dataset not found')
        dataset_num = dataset['dataset']

        passkey = self.auth_access_token

        jobs = await self.rest_client.request('GET','/datasets/{}/job_counts/status'.format(dataset_id))
        tasks = await self.rest_client.request('GET','/datasets/{}/task_counts/status'.format(dataset_id))
        task_info = await self.rest_client.request('GET','/datasets/{}/task_counts/name_status'.format(dataset_id))
        task_stats = await self.rest_client.request('GET','/datasets/{}/task_stats'.format(dataset_id))
        try:
            config = await self.rest_client.request('GET','/config/{}'.format(dataset_id))
        except Exception:
            config = {}
        for t in task_info:
            logger.info('task_info[%s] = %r', t, task_info[t])
            type_ = 'UNK'
            for task in config.get('tasks', []):
                if 'name' in task and task['name'] == t:
                    type_ = 'GPU' if 'requirements' in task and 'gpu' in task['requirements'] and task['requirements']['gpu'] else 'CPU'
                    break
            task_info[t] = {
                'name': t,
                'type': type_,
                'idle': task_info[t].get('idle', 0),
                'waiting': task_info[t].get('waiting', 0),
                'queued': task_info[t].get('queued', 0),
                'running': task_info[t].get('processing', 0),
                'complete': task_info[t].get('complete', 0),
                'error': task_info[t].get('failed', 0) + task_info[t].get('suspended', 0),
            }
        self.render('dataset_detail.html', dataset_id=dataset_id, dataset_num=dataset_num,
                    dataset=dataset, jobs=jobs, tasks=tasks, task_info=task_info,
                    task_stats=task_stats, passkey=passkey)


class TaskBrowse(PublicHandler):
    """Handle /task urls"""
    @authenticated
    async def get(self, dataset_id):
        self.statsd.incr('task_browse')
        status = self.get_argument('status',default=None)

        if status:
            tasks = await self.rest_client.request('GET','/datasets/{}/tasks?status={}'.format(dataset_id,status))
            for t in tasks:
                if 'job_index' not in tasks[t]:
                    job = await self.rest_client.request('GET', '/datasets/{}/jobs/{}'.format(dataset_id, tasks[t]['job_id']))
                    tasks[t]['job_index'] = job['job_index']
            passkey = self.auth_access_token
            self.render('task_browse.html',tasks=tasks, passkey=passkey)
        else:
            status = await self.rest_client.request('GET','/datasets/{}/task_counts/status'.format(dataset_id))
            self.render('tasks.html',status=status)


class Task(PublicHandler):
    """Handle /task urls"""
    @authenticated
    async def get(self, dataset_id, task_id):
        self.statsd.incr('task')
        status = self.get_argument('status', default=None)

        passkey = self.auth_access_token

        dataset = await self.rest_client.request('GET', '/datasets/{}'.format(dataset_id))
        task_details = await self.rest_client.request('GET','/datasets/{}/tasks/{}?status={}'.format(dataset_id, task_id, status))
        task_stats = await self.rest_client.request('GET','/datasets/{}/tasks/{}/task_stats?last=true'.format(dataset_id, task_id))
        if task_stats:
            task_stats = list(task_stats.values())[0]
        try:
            ret = await self.rest_client.request('GET','/datasets/{}/tasks/{}/logs?group=true'.format(dataset_id, task_id))
            logs = ret['logs']
            # logger.info("logs: %r", logs)
            ret2 = await self.rest_client.request('GET','/datasets/{}/tasks/{}/logs?keys=log_id|name|timestamp'.format(dataset_id, task_id))
            logs2 = ret2['logs']
            logger.info("logs2: %r", logs2)
            log_by_name = defaultdict(list)
            for log in sorted(logs2,key=lambda lg:lg['timestamp'] if 'timestamp' in lg else '',reverse=True):
                log_by_name[log['name']].append(log)
            for log in logs:
                if 'data' not in log or not log['data']:
                    log['data'] = ''
                log_by_name[log['name']][0] = log
        except Exception:
            log_by_name = {}
        self.render('task_detail.html', dataset=dataset, task=task_details, task_stats=task_stats, logs=log_by_name, passkey=passkey)


class JobBrowse(PublicHandler):
    """Handle /job urls"""
    @authenticated
    async def get(self, dataset_id):
        self.statsd.incr('job')
        status = self.get_argument('status',default=None)

        passkey = self.auth_access_token

        jobs = await self.rest_client.request('GET', '/datasets/{}/jobs'.format(dataset_id))
        if status:
            for t in list(jobs):
                if jobs[t]['status'] != status:
                    del jobs[t]
                    continue
        self.render('job_browse.html', jobs=jobs, passkey=passkey)


class Job(PublicHandler):
    """Handle /job urls"""
    @authenticated
    async def get(self, dataset_id, job_id):
        self.statsd.incr('job')
        status = self.get_argument('status',default=None)

        passkey = self.auth_access_token

        dataset = await self.rest_client.request('GET', '/datasets/{}'.format(dataset_id))
        job = await self.rest_client.request('GET', '/datasets/{}/jobs/{}'.format(dataset_id,job_id))
        args = {'job_id': job_id}
        if status:
            args['status'] = status
        tasks = await self.rest_client.request('GET', f'/datasets/{dataset_id}/tasks', args)
        job['tasks'] = list(tasks.values())
        job['tasks'].sort(key=lambda x:x['task_index'])
        self.render('job_detail.html', dataset=dataset, job=job, passkey=passkey)


class Documentation(PublicHandler):
    @catch_error
    async def get(self, url):
        # try to get the user, if available
        await self.get_current_user_async()
        self.statsd.incr('documentation')
        doc_path = str(importlib.resources.files('iceprod.server')/'data'/'docs')
        full_path = os.path.join(doc_path, url)
        if not full_path.startswith(doc_path):
            self.set_status(404)
            self.render('404.html', path='bad docs path')
            return
        self.write(documentation.load_doc(full_path))
        self.flush()


class Log(PublicHandler):
    @authenticated
    async def get(self, dataset_id, log_id):
        self.statsd.incr('log')
        ret = await self.rest_client.request('GET','/datasets/{}/logs/{}'.format(dataset_id, log_id))
        log_text = ret['data']
        html = '<html><head><title>' + ret['name'] + '</title></head><body>'
        html += log_text.replace('\n', '<br/>')
        html += '</body></html>'
        self.write(html)
        self.flush()


class Help(PublicHandler):
    """Help Page"""
    @catch_error
    async def get(self):
        # try to get the user, if available
        await self.get_current_user_async()
        self.statsd.incr('help')
        self.render('help.html')


class Other(PublicHandler):
    """Handle any other urls - this is basically all 404"""
    @catch_error
    async def get(self):
        # try to get the user, if available
        await self.get_current_user_async()
        self.statsd.incr('other')
        path = self.request.path
        self.set_status(404)
        self.render('404.html', path=path)


class Profile(PublicHandler):
    """Handle user profile page"""
    @authenticated
    async def get(self):
        self.statsd.incr('profile')
        username = self.current_user
        groups = self.auth_groups
        group_creds = {}
        for g in groups:
            if g != 'users':
                ret = await self.cred_rest_client.request('GET', f'/groups/{g}/credentials')
                group_creds[g] = ret
        user_creds = await self.cred_rest_client.request('GET', f'/users/{username}/credentials')
        self.render('profile.html', username=username, groups=groups,
                    group_creds=group_creds, user_creds=user_creds)

    @authenticated
    async def post(self):
        username = self.current_user

        if self.get_argument('add_icecube_token', ''):
            args = {
                'url': 'https://data.icecube.aq',
                'type': 'oauth',
                'access_token': self.auth_access_token,
            }
            if self.auth_refresh_token:
                args['refresh_token'] = self.auth_refresh_token
                args['expiration'] = (datetime.utcnow() + timedelta(days=30)).isoformat()
            await self.cred_rest_client.request('POST', f'/users/{username}/credentials', args)

        else:
            type_ = self.get_argument('type')
            args = {
                'url': self.get_argument('url'),
                'type': type_,
            }
            if type_ == 's3':
                args['buckets'] = [x for x in self.get_argument('buckets').split('\n') if x]
                args['access_key'] = self.get_argument('access_key')
                args['secret_key'] = self.get_argument('secret_key')
            elif type_ == 'oauth':
                if acc := self.get_argument('refresh_token', ''):
                    args['refresh_token'] = acc
                if ref := self.get_argument('refresh_token', ''):
                    args['refresh_token'] = ref
                if not (acc or ref):
                    raise tornado.web.HTTPError(400, reason='need access or refresh token')
            else:
                raise tornado.web.HTTPError(400, reason='bad cred type')

            if self.get_argument('add_user_cred', ''):
                await self.cred_rest_client.request('POST', f'/users/{username}/credentials', args)
            elif self.get_argument('add_group_cred', ''):
                groupname = self.get_argument('group')
                if groupname not in self.auth_groups or groupname == 'users':
                    raise tornado.web.HTTPError(400, reason='bad group name')
                await self.cred_rest_client.request('POST', f'/groups/{groupname}/credentials', args)

        # now show the profile page
        await self.get()


class Logout(PublicHandler):
    @catch_error
    async def get(self):
        self.statsd.incr('logout')
        self.clear_tokens()
        self.current_user = None
        self.request.uri = '/'  # for login redirect, fake the main page
        self.render('logout.html', status=None)


class HealthHandler(PublicHandler):
    """
    Handle health requests.
    """
    async def get(self):
        """
        Get health status.

        Returns based on exit code, 200 = ok, 400 = failure
        """
        status = {
            'now': nowstr(),
        }

        try:
            await self.system_rest_client.request('GET', '/dataset_summaries/status')
            status['rest_api'] = 'OK'
        except Exception:
            logger.info('error from REST API', exc_info=True)
            self.send_error(500, reason='error from REST API')
            return

        try:
            await self.cred_rest_client.request('GET', '/healthz')
            status['cred_service'] = 'OK'
        except Exception:
            logger.info('error from REST API', exc_info=True)
            self.send_error(500, reason='error from credential service')
            return

        self.write(status)


class Server:
    def __init__(self):
        default_config = {
            'HOST': 'localhost',
            'PORT': 8080,
            'DEBUG': False,
            'OPENID_URL': '',
            'OPENID_AUDIENCE': '',
            'ICEPROD_WEB_URL': 'https://iceprod2.icecube.wisc.edu',
            'ICEPROD_API_ADDRESS': 'https://iceprod2-api.icecube.wisc.edu',
            'ICEPROD_API_CLIENT_ID': '',
            'ICEPROD_API_CLIENT_SECRET': '',
            'ICEPROD_CRED_ADDRESS': 'https://credentials.iceprod.icecube.aq',
            'ICEPROD_CRED_CLIENT_ID': '',
            'ICEPROD_CRED_CLIENT_SECRET': '',
            'COOKIE_SECRET': '',
            'STATSD_ADDRESS': '',
            'STATSD_PREFIX': 'rest_api',
            'CI_TESTING': '',
        }
        config = from_environment(default_config)

        # get package data
        static_path = str(importlib.resources.files('iceprod.website')/'data'/'www')
        if static_path is None or not os.path.exists(static_path):
            logger.info('static path: %r',static_path)
            raise Exception('bad static path')
        template_path = str(importlib.resources.files('iceprod.website')/'data'/'www_templates')
        if template_path is None or not os.path.exists(template_path):
            logger.info('template path: %r',template_path)
            raise Exception('bad template path')

        # set IceProd REST API
        if config['ICEPROD_API_ADDRESS']:
            rest_address = config['ICEPROD_API_ADDRESS']
        else:
            raise RuntimeError('ICEPROD_API_ADDRESS not specified')

        rest_config = {
            'debug': config['DEBUG'],
            'server_header': 'IceProd/' + iceprod.__version__,
        }

        if config['OPENID_URL']:
            logging.info(f'enabling auth via {config["OPENID_URL"]} for aud "{config["OPENID_AUDIENCE"]}"')
            rest_config.update({
                'auth': {
                    'openid_url': config['OPENID_URL'],
                    'audience': config['OPENID_AUDIENCE'],
                }
            })
        elif config['CI_TESTING']:
            rest_config.update({
                'auth': {
                    'secret': 'secret',
                }
            })
        else:
            raise RuntimeError('OPENID_URL not specified, and CI_TESTING not enabled!')

        statsd = FakeStatsClient()
        if config['STATSD_ADDRESS']:
            try:
                addr = config['STATSD_ADDRESS']
                port = 8125
                if ':' in addr:
                    addr,port = addr.split(':')
                    port = int(port)
                statsd = StatsClientIgnoreErrors(addr, port=port, prefix=config['STATSD_PREFIX'])
            except Exception:
                logger.warning('failed to connect to statsd: %r', config['STATSD_ADDRESS'], exc_info=True)

        if config['ICEPROD_CRED_CLIENT_ID'] and config['ICEPROD_CRED_CLIENT_SECRET']:
            logging.info(f'enabling auth via {config["OPENID_URL"]} for aud "{config["OPENID_AUDIENCE"]}"')
            cred_client = ClientCredentialsAuth(
                address=config['ICEPROD_CRED_ADDRESS'],
                token_url=config['OPENID_URL'],
                client_id=config['ICEPROD_CRED_CLIENT_ID'],
                client_secret=config['ICEPROD_CRED_CLIENT_SECRET'],
            )
        elif config['CI_TESTING']:
            cred_client = RestClient(config['ICEPROD_CRED_ADDRESS'], timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_CRED_CLIENT_ID or ICEPROD_CRED_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        handler_args = RestHandlerSetup(rest_config)
        handler_args['cred_rest_client'] = cred_client

        full_url = config['ICEPROD_WEB_URL']
        handler_args['full_url'] = full_url
        login_url = full_url+'/login'

        login_handler_args = handler_args.copy()
        if config['ICEPROD_API_CLIENT_ID'] and config['ICEPROD_API_CLIENT_SECRET']:
            logging.info('enabling system rest client and website logins"')
            rest_client = ClientCredentialsAuth(
                address=config['ICEPROD_API_ADDRESS'],
                token_url=config['OPENID_URL'],
                client_id=config['ICEPROD_API_CLIENT_ID'],
                client_secret=config['ICEPROD_API_CLIENT_SECRET'],
            )
            login_handler_args['oauth_client_id'] = config['ICEPROD_API_CLIENT_ID']
            login_handler_args['oauth_client_secret'] = config['ICEPROD_API_CLIENT_SECRET']
            login_handler_args['oauth_client_scope'] = 'offline_access posix profile'
        elif config['CI_TESTING']:
            logger.info('CI_TESTING: no login for testing')
            rest_client = RestClient(config['ICEPROD_API_ADDRESS'], timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_API_CLIENT_ID or ICEPROD_API_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        handler_args.update({
            'statsd': statsd,
            'rest_api': rest_address,
            'system_rest_client': rest_client,
        })
        if config['COOKIE_SECRET']:
            cookie_secret = config['COOKIE_SECRET']
            log_cookie_secret = cookie_secret[:4] + 'X'*(len(cookie_secret)-8) + cookie_secret[-4:]
            logger.info('using supplied cookie secret %r', log_cookie_secret)
        else:
            cookie_secret = ''.join(hex(random.randint(0,15))[-1] for _ in range(64))

        server = RestServer(
            debug=config['DEBUG'],
            cookie_secret=cookie_secret,
            login_url=login_url,
            template_path=template_path,
            static_path=static_path,
        )

        server.add_route(r"/", Default, handler_args)
        server.add_route(r"/submit", Submit, handler_args)
        server.add_route(r"/config", Config, handler_args)
        server.add_route(r"/schemas/v3/([\w\.]+)", Schemas, handler_args)
        server.add_route(r"/dataset", DatasetBrowse, handler_args)
        server.add_route(r"/dataset/(\w+)", Dataset, handler_args)
        server.add_route(r"/dataset/(\w+)/task", TaskBrowse, handler_args)
        server.add_route(r"/dataset/(\w+)/task/(\w+)", Task, handler_args)
        server.add_route(r"/dataset/(\w+)/job", JobBrowse, handler_args)
        server.add_route(r"/dataset/(\w+)/job/(\w+)", Job, handler_args)
        server.add_route(r"/help", Help, handler_args)
        server.add_route(r"/docs/(.*)", Documentation, handler_args)
        server.add_route(r"/dataset/(\w+)/log/(\w+)", Log, handler_args)
        server.add_route(r'/profile', Profile, handler_args)
        server.add_route(r"/login", Login, login_handler_args)
        server.add_route(r"/logout", Logout, handler_args)
        server.add_route('/healthz', HealthHandler, handler_args)
        server.add_route(r"/.*", Other, handler_args)

        server.startup(address=config['HOST'], port=config['PORT'])

        self.server = server

    async def start(self):
        pass

    async def stop(self):
        await self.server.stop()



import json
import logging
import uuid

import tornado.web

from ..base_handler import APIBase
from ..auth import authorization, attr_auth
from iceprod.server.util import nowstr

logger = logging.getLogger('rest.logs')


def setup(handler_cfg):
    """
    Setup method for Logs REST API.

    Args:
        handler_cfg (dict): args to pass to the route

    Returns:
        dict: routes, database, indexes
    """
    return {
        'routes': [
            (r'/logs', MultiLogsHandler, handler_cfg),
            (r'/logs/(?P<log_id>\w+)', LogsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/logs', DatasetMultiLogsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/logs/(?P<log_id>\w+)', DatasetLogsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_id>\w+)/logs', DatasetTaskLogsHandler, handler_cfg),
        ],
        'database': 'logs',
        'indexes': {
            'logs': {
                'log_id_index': {'keys': 'log_id', 'unique': True},
                'task_id_index': {'keys': 'task_id', 'unique': False},
                'dataset_id_index': {'keys': 'dataset_id', 'unique': False},
            }
        }
    }


class MultiLogsHandler(APIBase):
    """
    Handle logs requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self):
        """
        Get multiple log entries based on search arguments.

        Body args (json):
            from (str): timestamp to start searching
            to (str): timestamp to end searching
            name (str): name (type) of log
            dataset_id (str): dataset_id
            task_id (str): task_id
            keys: | separated list of keys to return for each task

        Returns:
            dict: {log_id: {keys}}
        """
        query = {}
        projection = {'_id': False, 'log_id': True}
        try:
            limit = int(self.get_argument('limit', 0))
        except Exception:
            raise tornado.web.HTTPError(500, reason='non-integer limit')

        date_to = self.get_argument('to', None)
        date_from = self.get_argument('from', None)
        if date_to and date_from:
            query['timestamp'] = {'$gte': date_from, '$lte': date_to}
        elif date_from:
            query['timestamp'] = {'$gte': date_from}
        elif date_to:
            query['timestamp'] = {'$lte': date_to}

        for k in ('name', 'dataset_id', 'task_id'):
            val = self.get_argument(k, None)
            if val:
                query[k] = val

        keys = self.get_argument('keys', 'name|data|timestamp|task_id|dataset_id')
        projection.update({x:True for x in keys.split('|') if x})

        logging.debug('query: %r', query)
        logging.debug('projection: %r', projection)

        ret = {}
        f = self.db.logs.find(query, projection=projection, limit=limit)
        async for row in f:
            if 'data' in projection and 'data' not in row:
                if self.s3:
                    row['data'] = await self.s3.get(row['log_id'])
                else:
                    raise tornado.web.HTTPError(500, reason='no data field and s3 disabled')
            ret[row['log_id']] = row
        self.write(ret)
        self.finish()

    @authorization(roles=['admin', 'system'])
    async def post(self):
        """
        Create a log entry.

        Body should contain the following fields:
            data: str

        Optional fields:
            dataset_id: str
            task_id: str
            name: str

        Returns:
            dict: {'result': <log_id>}
        """
        data = json.loads(self.request.body)
        if 'data' not in data:
            raise tornado.web.HTTPError(400, reason='data field not in body')
        if 'name' not in data:
            data['name'] = 'log'
        log_id = uuid.uuid1().hex
        data['log_id'] = log_id
        data['timestamp'] = nowstr()
        if self.s3 and len(data['data']) > 1000000:
            await self.s3.put(log_id, data['data'])
            del data['data']
        await self.db.logs.insert_one(data)
        self.set_status(201)
        self.write({'result': log_id})
        self.finish()


class LogsHandler(APIBase):
    """
    Handle logs requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, log_id):
        """
        Get a log entry.

        Args:
            log_id (str): the log id of the entry

        Returns:
            dict: all body fields
        """
        ret = await self.db.logs.find_one({'log_id':log_id}, projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Log not found")
        else:
            if 'data' not in ret:
                if self.s3:
                    ret['data'] = await self.s3.get(ret['log_id'])
                else:
                    raise tornado.web.HTTPError(500, reason='no data field and s3 disabled')
            self.write(ret)
            self.finish()

    @authorization(roles=['admin', 'system'])
    async def delete(self, log_id):
        """
        Delete a log entry.

        Args:
            log_id (str): the log id of the entry

        Returns:
            dict: empty dict on success
        """
        ret = await self.db.logs.find_one_and_delete({'log_id':log_id})
        if ret:
            if 'data' not in ret:
                if self.s3:
                    e = await self.s3.exists(log_id)
                    if e:
                        await self.s3.delete(log_id)
                else:
                    logging.warn('no data field and s3 disabled')
        self.write({})
        self.finish()


class DatasetMultiLogsHandler(APIBase):
    """
    Handle logs requests.
    """
    @authorization(roles=['admin', 'user'])
    @attr_auth(arg='dataset_id', role='write')
    async def post(self, dataset_id):
        """
        Create a log entry.

        Body should contain the following fields:
            data: str

        Optional fields:
            task_id: str
            name: str

        Args:
            dataset_id (str): the dataset id

        Returns:
            dict: {'result': <log_id>}
        """
        data = json.loads(self.request.body)
        if 'data' not in data:
            raise tornado.web.HTTPError(400, reason='data field not in body')
        log_id = uuid.uuid1().hex
        data['log_id'] = log_id
        data['dataset_id'] = dataset_id
        data['timestamp'] = nowstr()
        if self.s3 and len(data['data']) > 1000000:
            await self.s3.put(log_id, data['data'])
            del data['data']
        await self.db.logs.insert_one(data)
        self.set_status(201)
        self.write({'result': log_id})
        self.finish()

    @authorization(roles=['admin', 'user'])
    @attr_auth(arg='dataset_id', role='write')
    async def delete(self, dataset_id, task_id):
        """
        Delete all logs for a dataset.

        Streams progress as deletion happens.

        Args:
            dataset_id (str): the dataset id
            task_id (str): the task id

        Returns:
            str: one json doc per line, {'status', 'percent_complete'}
        """
        total_docs = await self.db.logs.count_documents({'dataset_id': dataset_id})
        ret = {'status': 'deleting', 'percent_complete': 0., 'total': total_docs}
        i = 0
        async for row in self.db.logs.find({'dataset_id': dataset_id}):
            log_id = row['log_id']
            if 'data' not in row:
                if self.s3:
                    e = await self.s3.exists(log_id)
                    if e:
                        await self.s3.delete(log_id)
                else:
                    logging.warn('no data field and s3 disabled')
            await self.db.logs.delete_one({'log_id':log_id})
            i += 1
            ret['percent_complete'] = 100.*i/total_docs
            self.write(ret)
        self.write({'status': 'complete', 'percent_complete': 100.})
        self.finish()


class DatasetLogsHandler(APIBase):
    """
    Handle logs requests.
    """
    @authorization(roles=['admin', 'user'])
    @attr_auth(arg='dataset_id', role='write')
    async def get(self, dataset_id, log_id):
        """
        Get a log.

        Args:
            dataset_id (str): the dataset id
            log_id (str): the log id of the entry

        Returns:
            dict: all body fields
        """
        ret = await self.db.logs.find_one(
            {'dataset_id':dataset_id,'log_id':log_id},
            projection={'_id':False}
        )
        if not ret:
            self.send_error(404, reason="Log not found")
        else:
            if 'data' not in ret:
                if self.s3:
                    ret['data'] = await self.s3.get(ret['log_id'])
                else:
                    raise tornado.web.HTTPError(500, reason='no data field and s3 disabled')
            self.write(ret)
            self.finish()


class DatasetTaskLogsHandler(APIBase):
    """
    Handle log requests for a task
    """
    @authorization(roles=['admin', 'user'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id, task_id):
        """
        Get logs for a dataset and task.

        Note: "num" and "group" are generally not used together.

        Params (optional):
            num (int): number of logs, or groups of logs, to return
            group {true, false}: group by log name
            order {asc, desc}: order by time
            keys: | separated list of keys to return for each log

        Args:
            dataset_id (str): the dataset id
            task_id (str): the task id

        Returns:
            dict: {'logs': [log entry dict, log entry dict]}
        """
        filters = {'dataset_id': dataset_id, 'task_id': task_id}

        num = self.get_argument('num', None)
        if num:
            try:
                num = int(num)
            except Exception:
                raise tornado.web.HTTPError(400, reason='bad num param. must be int')

        group = self.get_argument('group', 'false').lower() == 'true'
        order = self.get_argument('order', 'desc').lower()
        if order not in ('asc', 'desc'):
            raise tornado.web.HTTPError(400, reason='bad order param. should be "asc" or "desc".')

        projection = {'_id': False}
        keys = self.get_argument('keys', None)
        if keys:
            projection.update({x:True for x in keys.split('|') if x})

        steps = [
            {'$match': filters},
            {'$sort': {'timestamp': -1 if order == 'desc' else 1}},
        ]
        if group:
            if not keys:
                keys = 'log_id|name|task_id|dataset_id|data|timestamp'
            grouping = {x:{'$first':'$'+x} for x in keys.split('|') if x}
            grouping['_id'] = '$name'
            if 'timestamp' not in grouping:
                grouping['timestamp'] = {'$first': '$timestamp'}
            steps.extend([
                {'$group': grouping},
                {'$sort': {'timestamp': -1 if order == 'desc' else 1}},
            ])
        steps.append({'$project': projection})
        if num:
            steps.append({'$limit': num})
        logger.debug('steps: %r', steps)

        cur = self.db.logs.aggregate(steps, allowDiskUse=True)
        ret = []
        async for entry in cur:
            ret.append(entry)
        if not ret:
            self.send_error(404, reason="Log not found")
        else:
            for log in ret:
                if 'data' not in log and ((not keys) or 'data' in keys.split('|')):
                    try:
                        if self.s3:
                            log['data'] = await self.s3.get(log['log_id'])
                        else:
                            raise Exception('no data field and s3 disabled')
                    except Exception as e:
                        self.send_error(500, reason=str(e))
                        return
            self.write({'logs':ret})
            self.finish()

    @authorization(roles=['admin', 'user'])
    @attr_auth(arg='dataset_id', role='write')
    async def delete(self, dataset_id, task_id):
        """
        Delete all logs for a dataset and task.

        Args:
            dataset_id (str): the dataset id
            task_id (str): the task id

        Returns:
            dict: empty dict on success
        """
        async for row in self.db.logs.find({'dataset_id': dataset_id, 'task_id': task_id}):
            log_id = row['log_id']
            if 'data' not in row:
                if self.s3:
                    e = await self.s3.exists(log_id)
                    if e:
                        await self.s3.delete(log_id)
                else:
                    logging.warn('no data field and s3 disabled')
            await self.db.logs.delete_one({'log_id':log_id})
        self.write({})
        self.finish()



"""
Clean up the logs based on retention policy.

iceprod_log: 1 month
stderr: 1 year
stdout: 1 year
"""

import argparse
import asyncio
from datetime import datetime, timedelta
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.server.util import datetime2str

logger = logging.getLogger('pilot_monitor')


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    async def delete_logs(name, days):
        time_limit = datetime.utcnow() - timedelta(days=days)
        args = {
            'to': datetime2str(time_limit),
            'name': name,
            'keys': 'log_id',
            'limit': 100,
        }
        logs = await rest_client.request('GET', '/logs', args)
        for log_id in logs:
            await rest_client.request('DELETE', '/logs/{}'.format(log_id))
        return len(logs)
    try:
        while (await delete_logs('stdlog', 31)) == 100:
            await asyncio.sleep(60)
        # while (await delete_logs('stderr', 365)) == 1000:
        #     await asyncio.sleep(60)
        # while (await delete_logs('stdout', 365)) == 1000:
        #     await asyncio.sleep(60)
    except Exception:
        logger.error('error cleaning logs', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')
    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""
Interface for configuring and submitting jobs on a computing cluster.
Do not use this class directly. Instead use one of the batch plugins
that inherit from this class.
"""

import asyncio
import os
import logging
from copy import deepcopy
from pathlib import Path
import socket
from typing import Any, Protocol

from asyncache import cachedmethod  # type: ignore
from cachetools import TTLCache
from cachetools.func import ttl_cache
import requests.exceptions

from iceprod.core import functions
from iceprod.core.config import Task, Job, Dataset
from iceprod.core.defaults import add_default_options
from iceprod.core.resources import Resources
from iceprod.server.states import JOB_STATUS_START
from iceprod.server.util import nowstr


logger = logging.getLogger('grid')


@ttl_cache(ttl=3600)
def get_host():
    """Cache the host fqdn for 1 hour"""
    return socket.getfqdn()


class GridTask(Protocol):
    """Protocol for grid task dataclass"""
    dataset_id: str | None
    task_id: str | None
    instance_id: str | None


class BaseGrid:
    """
    Interface for a generic job distribution system.
    Do not use this class directly.  Use one of the plugins.
    """

    def __init__(self, cfg, rest_client, cred_client):
        self.cfg = cfg
        self.rest_client = rest_client
        self.cred_client = cred_client

        queue_cfg = self.cfg['queue']

        # site name
        self.site = None
        if 'site' in queue_cfg:
            self.site = queue_cfg['site']

        # task queue params
        self.site_requirements = queue_cfg['resources']
        self.site_query_params = {}
        if self.site:
            self.site_requirements['site'] = self.site
            if queue_cfg.get('exclusive', False):
                logger.info('site requirements: must match site name')
                self.site_query_params['requirements.site'] = self.site

            if 'gpu' in self.site.lower():
                logger.info('site requirements: GPU site!')
                self.site_requirements['gpu'] = 1

        # directories
        self.credentials_dir = Path(os.path.expanduser(os.path.expandvars(
            queue_cfg['credentials_dir'])))
        self.credentials_dir.mkdir(mode=0o700, parents=True, exist_ok=True)

        self.submit_dir = Path(os.path.expanduser(os.path.expandvars(
            queue_cfg['submit_dir'])))
        self.submit_dir.mkdir(mode=0o700, parents=True, exist_ok=True)

        # dataset lookup cache
        self.dataset_cache = TTLCache(maxsize=100, ttl=60)

    # Functions to override #

    async def run(self):
        """Override the `ActiveJobs` and `JobActions` batch submission / monitoring classes"""
        raise NotImplementedError()

    # Commong functions #

    @cachedmethod(lambda self: self.dataset_cache)
    async def dataset_lookup(self, dataset_id: str) -> Dataset:
        """
        Lookup a dataset fromn the REST API.

        Uses caching with TTL 60 seconds.

        Args:
            dataset_id: dataset id

        Returns:
            iceprod.core.config.Dataset object
        """
        ret = await Dataset.load_from_api(dataset_id, rest_client=self.rest_client)
        ret.fill_defaults()
        ret.validate()
        return ret

    async def get_tasks_to_queue(self, num: int) -> list[Task]:
        """
        Get new tasks to queue from the REST API.

        Args:
            num: number of tasks to retrieve

        Returns:
            list of iceprod.core.config.Task objects
        """
        # get tasks to run from REST API, and convert to batch jobs
        args = {
            'requirements': self.site_requirements,
            'query_params': self.site_query_params,
        }
        futures = set()
        tasks_queued = 0
        for tasks_queued in range(num):
            try:
                ret = await self.rest_client.request('POST', '/task_actions/queue', args)
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404:
                    break
                raise
            else:
                futures.add(asyncio.create_task(self._convert_to_task(ret)))
        logging.info('got %d tasks to queue', tasks_queued)

        tasks = []
        for f in asyncio.as_completed(futures):
            task = await f
            # add default resource requirements
            task.requirements = self._get_resources(task)
            tasks.append(task)

        return tasks

    async def _convert_to_task(self, task):
        """Convert from basic task dict to a Task object"""
        d = deepcopy(await self.dataset_lookup(task['dataset_id']))
        # don't bother looking up the job status - trust that if we got a task, we're in processing
        j = Job(dataset=d, job_id=task['job_id'], job_index=task['job_index'], status=JOB_STATUS_START)
        t = Task(
            dataset=d,
            job=j,
            task_id=task['task_id'],
            task_index=task['task_index'],
            instance_id=task['instance_id'],
            name=task['name'],
            depends=task['depends'],
            requirements=task['requirements'],
            status=task['status'],
            site=self.site,
            stats={},
            task_files=[],
        )
        await t.load_task_files_from_api(self.rest_client)

        # load some config defaults
        config = t.dataset.config
        if (not config['options'].get('site_temp','')) and self.cfg['queue'].get('site_temp', ''):
            config['options']['site_temp'] = self.cfg['queue']['site_temp']
        add_default_options(config['options'])

        return t

    @staticmethod
    def _get_resources(task):
        """
        Combine task resources with default resources.

        Args:
            task (Task): IceProd Task
        Returns:
            dict: resources
        """
        resource = deepcopy(Resources.defaults)
        for k in resource:
            if isinstance(resource[k],list):
                resource[k] = len(resource[k])
        values = {}
        try:
            for k in task.requirements:
                if k in resource and task.requirements[k]:
                    try:
                        if isinstance(resource[k], int):
                            values[k] = int(task.requirements[k])
                        elif isinstance(resource[k], float):
                            values[k] = float(task.requirements[k])
                        else:
                            values[k] = task.requirements[k]
                    except Exception:
                        logger.warning('bad reqs value for task %r', task)
                elif k == 'os' and task.requirements[k]:
                    logger.debug('OS req: %s', task.requirements[k])
                    values['os'] = task.requirements[k] if isinstance(task.requirements[k], list) else [task.requirements[k]]
        except TypeError:
            logger.warning('task.requirements: %r', task.requirements)
            raise
        resource.update(values)
        return resource

    # Task Actions #

    async def _upload_log(self, task: GridTask, name: str, data: str):
        """
        Upload a log to the IceProd API.

        Args:
            task: IceProd task info
            name: log name
            data: log text data
        """
        args = {
            'dataset_id': task.dataset_id,
            'task_id': task.task_id,
            'name': name,
            'data': data,
        }
        try:
            await self.rest_client.request('POST', '/logs', args)
        except requests.exceptions.HTTPError:
            logger.warning('cannot upload log', exc_info=True)

    async def _upload_stats(self, task: GridTask, stats: dict):
        """
        Upload task statistics to the IceProd API.

        Args:
            task: IceProd task info
            stats: stats dict
        """
        stats_cp = stats.copy()

        hostname = functions.gethostname()
        domain = '.'.join(hostname.split('.')[-2:])
        site = self.site
        if 'site' in stats_cp:
            site = stats_cp.pop('site', '')
        args = {
            'dataset_id': task.dataset_id,
            'hostname': hostname,
            'domain': domain,
            'site': site,
            'resources': stats_cp.pop('resources', {}),
            'task_stats': stats_cp,
            'time': nowstr(),
        }
        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_stats', args)
        except requests.exceptions.HTTPError:
            logger.warning('cannot upload stats', exc_info=True)

    async def task_idle(self, task: GridTask):
        """
        Tell IceProd API a task is now idle on the queue (put back to "queue" status).

        Args:
            task: IceProd task info
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        args = {
            'status': 'queued',
            'instance_id': task.instance_id,
        }
        try:
            await self.rest_client.request('PUT', f'/tasks/{task.task_id}/status', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise

    async def task_processing(self, task: GridTask):
        """
        Tell IceProd API a task is now processing (put in the "processing" status).

        Args:
            task: IceProd task info
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        args = {
            'instance_id': task.instance_id,
        }
        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_actions/processing', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise

    async def task_reset(self, task: GridTask, reason: str | None = None):
        """
        Tell IceProd API a task should be reset back to the "waiting" status.

        Args:
            task: IceProd task info
            reason: A reason for failure
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        args = {
            'instance_id': task.instance_id,
        }
        if reason:
            args['reason'] = reason
        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_actions/reset', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise
        else:
            if reason:
                await self._upload_log(task, 'stdlog', reason)

    async def task_failure(self, task: GridTask, reason: str | None = None, stats: dict | None = None, stdout: Path | None = None, stderr: Path | None = None):
        """
        Tell IceProd API a task should be put in the "failed" status.

        Stats format:
            {
                resources: {cpu: 1, gpu: 1, ...},
                ...
            }

        Args:
            task: IceProd task info
            reason: A reason for failure
            stats: task resource statistics
            stdout: path to stdout file
            stderr: path to stderr file
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        args: dict[str, Any] = {
            'instance_id': task.instance_id,
        }
        if reason:
            args['reason'] = reason
        if stats:
            args['resources'] = stats.get('resources', {})
        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_actions/failed', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise
        else:
            if stats:
                await self._upload_stats(task, stats)
            if stdout and stdout.exists():
                await self._upload_log(task, 'stdout', stdout.read_text())
            if stderr and stderr.exists():
                await self._upload_log(task, 'stderr', stderr.read_text())
            if reason:
                await self._upload_log(task, 'stdlog', reason)

    async def task_success(self, task: GridTask, stats: dict | None = None, stdout: Path | None = None, stderr: Path | None = None):
        """
        Tell IceProd API a task was successfully completed.

        Args:
            task: IceProd task info
            stats: task resource statistics
            stdout: path to stdout file
            stderr: path to stderr file
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        site = self.site
        if stats and 'site' in stats:
            site = stats['site']
        args = {
            'site': site,
            'instance_id': task.instance_id,
        }
        if stats:
            resources = stats.get('resources', {})
            if 'time' in resources:
                args['time_used'] = resources['time']*3600.

        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_actions/complete', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise
        else:
            if stats:
                await self._upload_stats(task, stats)
            if stdout and stdout.exists():
                await self._upload_log(task, 'stdout', stdout.read_text())
            if stderr and stderr.exists():
                await self._upload_log(task, 'stderr', stderr.read_text())


'''
    @run_on_executor
    def _delete_dirs(self, dirs):
        # delete dirs that need deleting
        for t in dirs:
            if not t.startswith(self.submit_dir):
                # some security against nefarious things
                raise Exception('directory %s not in submit_dir %s'%(t, self.submit_dir))
            try:
                logger.info('deleting submit_dir %s', t)
                functions.removedirs(t)
            except Exception:
                logger.warning('could not delete submit dir %s', t, exc_info=True)
                continue


    @ttl_cache(ttl=600)
    def get_token(self):
        return self.cred_client.make_access_token()

    @cached(TTLCache(1024, 60))
    async def get_user_credentials(self, username):
        ret = await self.cred_client.request('GET', f'/users/{username}/credentials')
        return ret

    @cached(TTLCache(1024, 60))
    async def get_group_credentials(self, group):
        ret = await self.cred_client.request('GET', f'/groups/{group}/credentials')
        return ret

    async def customize_task_config(self, task_cfg, job_cfg=None, dataset=None):
        """Transforms for the task config"""
        logger.info('customize_task_config for %s', job_cfg.get('options', {}).get('task_id', 'unknown'))

        # first expand site temp urls
        def expand_remote(cfg):
            new_data = []
            for d in cfg.get('data', []):
                if not d['remote']:
                    try:
                        remote_base = d.storage_location(job_cfg)
                        logger.info('expanding remote for %r', d['local'])
                        d['remote'] = os.path.join(remote_base, d['local'])
                    except Exception:
                        # ignore failed expansions, as these are likely local temp paths
                        pass
                new_data.append(d)
            cfg['data'] = new_data
        expand_remote(task_cfg)
        for tray in task_cfg['trays']:
            expand_remote(tray)
            for module in tray['modules']:
                expand_remote(module)
        logger.info('task_cfg: %r', task_cfg)

        # now apply S3 and token credentials
        creds = self.cfg.get('creds', {})
        if dataset['group'] == 'users':
            ret = await self.get_user_credentials(dataset['username'])
        else:
            ret = await self.get_group_credentials(dataset['group'])
        creds.update(ret)

        s3_creds = {url: creds.pop(url) for url in list(creds) if creds[url]['type'] == 's3'}
        if s3_creds:
            # if we have any s3 credentials, try presigning urls
            logger.info('testing job for s3 credentials')
            try:
                queued_time = timedelta(seconds=self.queue_cfg['max_task_queued_time'])
            except Exception:
                queued_time = timedelta(seconds=86400*2)
            try:
                processing_time = timedelta(seconds=self.queue_cfg['max_task_processing_time'])
            except Exception:
                processing_time = timedelta(seconds=86400*2)
            expiration = (queued_time + processing_time).total_seconds()
            logger.info(f's3 cred expire time: {expiration}')

            def presign_s3(cfg):
                new_data = []
                for d in cfg.get('data', []):
                    for url in s3_creds:
                        if d['remote'].startswith(url):
                            logger.info('found data for cred: %s', url)
                            path = d['remote'][len(url):].lstrip('/')
                            bucket = None
                            if '/' in path:
                                bucket, key = path.split('/', 1)
                            if (not bucket) or bucket not in s3_creds[url]['buckets']:
                                key = path
                                bucket = urlparse(url).hostname.split('.', 1)[0]
                                if bucket not in s3_creds[url]['buckets']:
                                    raise RuntimeError('bad s3 bucket')

                            while '//' in key:
                                key = key.replace('//', '/')
                            while key.startswith('/'):
                                key = key[1:]

                            s = S3(url, s3_creds[url]['access_key'], s3_creds[url]['secret_key'], bucket=bucket)
                            logger.info(f'S3 url={url} bucket={bucket} key={key}')
                            if d['movement'] == 'input':
                                d['remote'] = s.get_presigned(key, expiration=expiration)
                                new_data.append(d)
                            elif d['movement'] == 'output':
                                d['remote'] = s.put_presigned(key, expiration=expiration)
                            elif d['movement'] == 'both':
                                d['movement'] = 'input'
                                d['remote'] = s.get_presigned(key, expiration=expiration)
                                new_data.append(d.copy())
                                d['movement'] = 'output'
                                d['remote'] = s.put_presigned(key, expiration=expiration)
                            else:
                                raise RuntimeError('unknown s3 data movement')
                            new_data.append(d)
                            break
                    else:
                        new_data.append(d)
                cfg['data'] = new_data

            presign_s3(task_cfg)
            for tray in task_cfg['trays']:
                presign_s3(tray)
                for module in tray['modules']:
                    presign_s3(module)

        oauth_creds = {url: creds.pop(url) for url in list(creds) if creds[url]['type'] == 'oauth'}
        if oauth_creds:
            # if we have token-based credentials, add them to the config
            logger.info('testing job for oauth credentials')
            cred_keys = set()

            def get_creds(cfg):
                for d in cfg.get('data', []):
                    for url in oauth_creds:
                        if d['remote'].startswith(url):
                            logger.info('found data for cred: %s', url)
                            cred_keys.add(url)
                            break

            get_creds(task_cfg)
            for tray in task_cfg['trays']:
                get_creds(tray)
                for module in tray['modules']:
                    get_creds(module)

            file_creds = {}
            for url in cred_keys:
                cred_name = hashlib.sha1(oauth_creds[url]['access_token'].encode('utf-8')).hexdigest()
                path = os.path.join(self.credentials_dir, cred_name)
                if not os.path.exists(path):
                    with open(path, 'w') as f:
                        f.write(oauth_creds[url]['access_token'])
                file_creds[url] = cred_name
            job_cfg['options']['credentials'] = file_creds

    async def setup_submit_directory(self,task):
        """Set up submit directory"""
        # create directory for task
        submit_dir = self.submit_dir
        task_dir = os.path.join(submit_dir,task['task_id']+'_'+str(random.randint(0,1000000)))
        while os.path.exists(task_dir):
            task_dir = os.path.join(submit_dir,task['task_id']+'_'+str(random.randint(0,1000000)))
        task_dir = os.path.abspath(os.path.expanduser(os.path.expandvars(task_dir)))
        os.makedirs(task_dir)
        task['submit_dir'] = task_dir

        # symlink or copy the .sh file
        src = get_pkg_binary('iceprod', 'loader.sh')
        dest = os.path.join(task_dir, 'loader.sh')
        try:
            os.symlink(src, dest)
        except Exception:
            try:
                functions.copy(src, dest)
            except Exception:
                logger.error('Error creating symlink or copy of .sh file: %s',dest,exc_info=True)
                raise

        # get passkey
        # expiration = self.queue_cfg['max_task_queued_time']
        # expiration += self.queue_cfg['max_task_processing_time']
        # expiration += self.queue_cfg['max_task_reset_time']
        # TODO: take expiration into account
        passkey = self.get_token()

        # write cfg
        cfg, filelist = self.write_cfg(task)

        # create submit file
        try:
            await asyncio.ensure_future(self.generate_submit_file(
                task,
                cfg=cfg,
                passkey=passkey,
                filelist=filelist
            ))
        except Exception:
            logger.error('Error generating submit file',exc_info=True)
            raise

    def write_cfg(self, task):
        """Write the config file for a task-like object"""
        filename = os.path.join(task['submit_dir'],'task.cfg')
        filelist = [filename]

        config = self.create_config(task)
        if creds := config['options'].get('credentials', {}):
            cred_dir = os.path.join(task['submit_dir'], CRED_SUBMIT_DIR)
            os.mkdir(cred_dir)
            for name in creds.values():
                src = os.path.join(self.credentials_dir, name)
                dest = os.path.join(cred_dir, name)
                os.symlink(src, dest)
            filelist.append(cred_dir)
        if 'system' in self.cfg and 'remote_cacert' in self.cfg['system']:
            config['options']['ssl'] = {}
            config['options']['ssl']['cacert'] = os.path.basename(self.cfg['system']['remote_cacert'])
            src = self.cfg['system']['remote_cacert']
            dest = os.path.join(task['submit_dir'],config['options']['ssl']['cacert'])
            try:
                os.symlink(src,dest)
            except Exception:
                try:
                    functions.copy(src,dest)
                except Exception:
                    logger.error('Error creating symlink or copy of remote_cacert',
                                 exc_info=True)
                    raise
            filelist.append(dest)
        if 'x509proxy' in self.cfg['queue'] and self.cfg['queue']['x509proxy']:
            config['options']['x509'] = os.path.basename(self.cfg['queue']['x509proxy'])
            src = self.cfg['queue']['x509proxy']
            logger.info('submit_dir %r  x509 %r', task['submit_dir'], config['options']['x509'])
            dest = os.path.join(task['submit_dir'],config['options']['x509'])
            try:
                os.symlink(src,dest)
            except Exception:
                try:
                    functions.copy(src,dest)
                except Exception:
                    logger.error('Error creating symlink or copy of x509 proxy',
                                 exc_info=True)
                    raise
            filelist.append(dest)
        if 'extra_file_tranfers' in self.cfg['queue'] and self.cfg['queue']['extra_file_tranfers']:
            for f in self.cfg['queue']['extra_file_tranfers']:
                logger.info('submit_dir %r  extra_files %r', task['submit_dir'], f)
                dest = os.path.join(task['submit_dir'],os.path.basename(f))
                try:
                    os.symlink(os.path.abspath(f),dest)
                except Exception:
                    try:
                        functions.copy(f,dest)
                    except Exception:
                        logger.error('Error creating symlink or copy of extra file %s',
                                     f, exc_info=True)
                        raise
                filelist.append(dest)
        if 'data_movement_stats' in self.cfg['queue'] and self.cfg['queue']['data_movement_stats']:
            config['options']['data_movement_stats'] = self.cfg['queue']['data_movement_stats']
        if 'upload_checksum' in self.cfg['queue']:
            config['options']['upload_checksum'] = self.cfg['queue']['upload_checksum']

        if 'reqs' in task:
            # add resources
            config['options']['resources'] = {}
            for r in task['reqs']:
                config['options']['resources'][r] = task['reqs'][r]

        # write to file
        serialization.serialize_json.dump(config, filename)

        c = iceprod.core.exe.Config(config)
        config = c.parseObject(config, {})

        return (config, filelist)

    # not async: called from executor
    def get_submit_args(self,task,cfg=None,passkey=None):
        """Get the submit arguments to start the loader script."""
        # get website address
        if ('rest_api' in self.cfg and self.cfg['rest_api'] and
                'url' in self.cfg['rest_api'] and self.cfg['rest_api']['url']):
            web_address = self.cfg['rest_api']['url']
        else:
            raise Exception('no web address for rest calls')

        args = []
        if 'software_dir' in self.queue_cfg and self.queue_cfg['software_dir']:
            args.append('-s {}'.format(self.queue_cfg['software_dir']))
        if 'iceprod_dir' in self.queue_cfg and self.queue_cfg['iceprod_dir']:
            args.append('-e {}'.format(self.queue_cfg['iceprod_dir']))
        if 'x509proxy' in self.cfg['queue'] and self.cfg['queue']['x509proxy']:
            args.append('-x {}'.format(os.path.basename(self.cfg['queue']['x509proxy'])))
        if ('download' in self.cfg and 'http_proxy' in self.cfg['download']
                and self.cfg['download']['http_proxy']):
            args.apend('-c {}'.format(self.cfg['download']['http_proxy']))
        args.append('--url {}'.format(web_address))
        if passkey:
            args.append('--passkey {}'.format(passkey))
        if cfg:
            args.append('--cfgfile task.cfg')
        if 'debug' in task and task['debug']:
            args.append('--debug')
        return args
'''
